type var = string

type exp =
  | Var of var
  | Add of (exp * exp)
  | Mul of (exp * exp)
  | Neg of exp
  | Const of int64

type cmd =
  | Skip
  | Assn    of var * exp
  | IfLZ    of exp * cmd * cmd
  | WhileLZ of exp * cmd
  | Seq     of cmd * cmd
  | Return  of exp

(* Interpreter -------------------------------------------------------------- *)

type state = var -> int64

let rec interpret_exp (s:state) (e:exp) : int64 =
  match e with
  | Var x -> s x
  | Add (e1, e2) -> Int64.add (interpret_exp s e1) (interpret_exp s e2)
  | Mul (e1, e2) -> Int64.mul (interpret_exp s e1) (interpret_exp s e2)
  | Const n -> n
  | Neg e -> Int64.neg (interpret_exp s e)

let update (s:state) (x:var) (v:int64) =
  fun y -> if x = y then v else s y

exception Halt of int64

let rec interpret_cmd (s:state) (c:cmd) : state =
  match c with
  | Skip -> s
  | Assn (x, e1) ->
    let v = interpret_exp s e1 in
    update s x v
  | IfLZ (e1, c1, c2) ->
    if Int64.compare (interpret_exp s e1) Int64.zero < 0
    then interpret_cmd s c1
    else interpret_cmd s c2
  | WhileLZ (e, c) ->
    if Int64.compare (interpret_exp s e) Int64.zero < 0 then
      interpret_cmd s (Seq(c, WhileLZ (e, c)))
    else
      s
  | Seq (c1, c2) ->
    let s1 = interpret_cmd s c1 in
    interpret_cmd s1 c2
  | Return e ->
    raise (Halt (interpret_exp s e))
  
let init_state : state = fun _ -> 0L

let answer program =
  try
    ignore (interpret_cmd init_state program "");
    0L;
  with Halt x -> x 

module OrderedVars = struct
  type t = var
  let compare = String.compare
end

module VSet = Set.Make(OrderedVars)
let (++) = VSet.union

(* 
  Calculate the set of variables mentioned in either an expression or a command. 
*)

let rec vars_of_exp (e:exp) : VSet.t =
  begin match e with
    | Var "x1" | Var "x2" | Var "x3" | Var "x4"
    | Var "x5" | Var "x6" | Var "x7" | Var "x8" ->
      VSet.empty
    | Var x -> VSet.singleton x
    | Add(e1, e2)
    | Mul(e1, e2) ->
      (vars_of_exp e1) ++ (vars_of_exp e2)
    | Neg(e) -> vars_of_exp e
    | Const _ -> VSet.empty
  end

let rec vars_of_cmd (c:cmd) : VSet.t =
  begin match c with
    | Skip -> VSet.empty
    | Assn(x, e) -> (VSet.singleton x) ++ (vars_of_exp e)
    | IfLZ(e, c1, c2) ->
      (vars_of_exp e) ++
      (vars_of_cmd c1) ++
      (vars_of_cmd c2)
    | WhileLZ(e, c) ->
      (vars_of_exp e) ++
      (vars_of_cmd c)
    | Seq(c1, c2) ->
      (vars_of_cmd c1) ++
      (vars_of_cmd c2)
    | Return e -> vars_of_exp e
  end

let var_list (c:cmd) : var list = VSet.elements (vars_of_cmd c)
